@@ -23,7 +23,7 @@ our @EXPORT = qw(
 	
 );
 
-our $VERSION = '0.09';
+our $VERSION = '0.10';
 
 
 # Preloaded methods go here.
@@ -88,6 +88,11 @@ sub match {
   return \@m;
 }
 
+sub get_env {
+  my $name = shift;
+  return $ENV{$name};
+}
+
 1;
 __END__
 
